__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.cache as cache
import citelang.main.settings as settings


def main(args, parser, extra, subparser):

    # init global settings
    cli = cache.cache

    if args.clear:
        cli.clear(force=args.force)
    else:
        print(settings.cfg.cache_dir)



#!/usr/bin/python

import os

import pytest

import citelang.main.client as client
import citelang.main.schemas as schemas
import citelang.utils as utils

here = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize(
    "name,filename,deps",
    [
        (
            "python-lib",
            "requirements.txt",
            [
                "sphinx",
            ],
        ),
        (
            "cpp-lib",
            "CMakeLists.txt",
            [
                "spack",
                "python",
                "curl",
                "xpat",
                "zlib",
                "CMakeLists.txt",
                "cpp-lib",
            ],
        ),
        (
            "r-lib",
            "DESCRIPTION",
            [
                "cran",
                "rmarkdown",
                "knitr",
                "callr",
                "methods",
                "R",
            ],
        ),
        (
            "python-lib",
            "setup.py",
            ["pypi", "pybind11", "types", "sphinx", "black"],
        ),
        #        (
        #            "go-lib",
        #            "go.mod",
        #            ["go-lib"],
        #        ),
        (
            "npm-lib",
            "package.json",
            ["gulp", "markdown", "mixin", "tape", "readable", "object"],
        ),
        (
            "gemfile-lib",
            "Gemfile",
            ["bundler", "rake", "jekyll", "rouge"],
        ),
    ],
)
def test_package_files(name, filename, deps):
    """
    Test loading custom package files
    """
    cli = client.get_parser(filename=os.path.join(here, "testdata", filename))
    result = cli.gen(name=name)
    content = result.render()
    for string in ["Software Credit", name, filename] + deps:
        assert string in content
    print(content)


@pytest.mark.parametrize(
    "manager,name",
    [
        ("pypi", "requests"),
        ("spack", "zlib"),
        #        ("github", "singularityhub/singularity-cli"),
    ],
)
def test_package(manager, name, tmp_path):
    """
    Test packages
    """
    cli = client.Client(quiet=False)
    result = cli.package(name=name, manager=manager)
    for prop in schemas.package["required"]:
        assert prop in result.data
    result.print_json()

    outfile = os.path.join(str(tmp_path), "%s.json" % os.path.basename(name))
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.table()


def test_list(tmp_path):
    """
    Test listing package managers
    """
    cli = client.Client(quiet=False)
    result = cli.package_managers()

    # 30 at time of parsing
    assert len(result.data) > 30
    result.print_json()

    outfile = os.path.join(str(tmp_path), "package_managers.json")
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.table()


@pytest.mark.parametrize(
    "manager,name",
    [
        ("pypi", "requests"),
        ("spack", "zlib"),
        #        ("github", "singularityhub/singularity-cli"),
    ],
)
def test_credit(manager, name, tmp_path):
    """
    Test credit function
    """
    cli = client.Client(quiet=False)
    result = cli.credit(name=name, manager=manager)
    result.print_json()

    outfile = os.path.join(str(tmp_path), "%s.json" % os.path.basename(name))
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.print_result()


@pytest.mark.parametrize("graphtype", [(None), ("cypher"), ("gexf"), ("dot")])
def test_graph(graphtype):
    cli = client.Client(quiet=False)
    cli.graph(name="requests", manager="pypi", fmt=graphtype)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import re

from citelang.logger import logger

from .base import PackagesFromFile


class RPackageManager(PackagesFromFile):
    """
    Packages from an R DESCRIPTION file, meaning an R package
    """

    name = "DESCRIPTION"
    underlying_manager = "cran"
    default_language = "R"
    project_count = None
    homepage = "https://cran.r-project.org/"
    color = "#006dad"
    default_versions = None

    def parse(self, content):
        """
        Parse the self.content (the DESCRIPTION file)
        """
        repo = self.get_repo()

        libs = []
        parsing = False
        for line in content.split("\n"):
            # If we are parsing, but hit the end of the list
            if parsing and not line.startswith(" "):
                break
            elif parsing:
                libs.append(line.replace(",", "").strip())
            elif "Imports:" in line:
                parsing = True

        deps = []
        for line in libs:
            version = None
            package_name = None
            if re.search("(==|<=|>=)", line):
                line = line.replace(")", "").replace("(", "")
                package_name, _, version = re.split("(==|<=|>=)", line)
                package_name = package_name.strip()
                version = version.strip()
            else:
                package_name = line

            # We cannot parse a dep without a name
            if not package_name:
                continue

            # Try to get from cache - either versioned or not
            pkg = self.get_package(package_name, version)
            if not pkg:
                logger.warning(f"Cannot find package {package_name}")
                continue

            # Ensure we have version, fallback to latest
            if not version:
                version = pkg.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/cran/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": pkg.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["Cran"],
            }
            deps.append(dep)

        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import sys

import citelang.defaults as defaults
import citelang.main.settings as settings
from citelang.logger import logger


def main(args, parser, extra, subparser):

    # If nothing provided, show help
    if not args.params:
        print(subparser.format_help())
        sys.exit(0)

    # The first "param" is either set of get
    command = args.params.pop(0)

    # init global settings
    validate = True if not command == "edit" else False

    # If the user wants the central config file
    if args.central:
        cfg = settings.SettingsBase(
            settings_file=defaults.default_settings_file, validate=validate
        )
    else:
        cfg = settings.SettingsBase(validate=validate)

    # For each new setting, update and save!
    if command == "inituser":
        return cfg.inituser()
    if command == "edit":
        return cfg.edit()

    if command in ["set", "add", "remove"]:

        # Update each param
        for param in args.params:
            cfg.update_param(command, param)

        # Save settings
        cfg.save()

    # For each get request, print the param pair
    elif command == "get":
        for key in args.params:
            value = cfg.get(key)
            value = value or "is unset"
            logger.info("%s %s" % (key.ljust(30), value))

    else:
        logger.error("%s is not a recognized command." % command)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import re

import citelang.defaults as defaults
from citelang.logger import logger

try:
    # The python manager that uses pip is more reliable
    if defaults.use_pip:
        from .pip_wrapper import PipManager as PythonManager
    else:
        from .python_base import PythonManager
except Exception:
    # vs. static parsing, not perfect but fairly good!
    from .python_base import PythonManager

    logger.warning(
        "pip not installed on system, parsing will be done statically (less reliable)"
    )


class RequirementsManager(PythonManager):

    """
    Packages parsed from a requirements.txt file (so from pypi)
    """

    name = "requirements.txt"
    underlying_manager = "pypi"
    project_count = None
    homepage = "pypi.org/"
    color = "#006dad"
    default_language = "Python"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Parse the self.content (the requirements.txt file)
        """
        repo = self.get_repo()
        lines = content.split("\n")
        deps = self.parse_python_deps(lines)
        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo


class SetupManager(PythonManager):

    """
    Packages parsed from a requirements.txt file (so from pypi)
    """

    name = "setup.py"
    underlying_manager = "pypi"
    project_count = None
    homepage = "pypi.org/"
    color = "#006dad"
    default_language = "Python"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Parse the self.content (the requirements.txt file)
        """
        repo = self.get_repo()

        lines = []
        parsing = False
        for line in content.split("\n"):

            if "setup_requires" in line or "install_requires" in line:
                line = line.replace(" ", "")

                # It MUST be a list
                if "=[" in line:
                    parsing = True

            # We found the start and end
            if parsing and "[" in line and "]" in line:
                lines.append(line)
                parsing = False

            elif parsing and "[" in line:
                lines.append(line)

            elif parsing and "]" in line:
                lines.append(line)
                parsing = False

            elif parsing:
                lines.append(line)

        # Always skip these, try to avoid Python iterations
        skips = ["@git+", "strip()", "^#", ".git" " else ", ".format", ".startswith"]
        skip_regex = "(%s)" % "|".join(skips)

        # Clean up lines
        cleaned = []
        for line in lines:

            # Remove line immediately if starts with comment
            if line.strip().startswith("#"):
                continue

            # Get rid of comments off the bat
            line = line.split("#", 1)[0]
            terms = ["setup_requires" + x for x in [":", " :", "=", " =", ""]]
            terms += ["install_requires" + x for x in [":", " :", "=", " =", ""]]
            for term in terms + ["[", "]", '"', "'", "+", "{", "}", "(", ")"]:
                line = line.replace(term, "")

            # Get rid of any sys_platform, etc.
            line = line.split(";")[0]
            parts = line.split(",")

            # Don't add any git pip installs
            cleaned += [
                x.strip()
                for x in parts
                if x.strip() and not re.search(skip_regex, x.strip())
            ]

        # Don't include any that don't have letters
        cleaned = [
            x
            for x in cleaned
            if re.search("[a-zA-Z]", x) and not re.search("^f('|\")", x)
        ]

        # Anything that starts with a comment
        cleaned = [x.strip() for x in cleaned if not x.strip().startswith("#")]

        # Remove any variants (e.g., [all])
        cleaned = [re.sub("\[.+\]", "", x) for x in cleaned]  # noqa
        deps = self.parse_python_deps(cleaned)
        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import os
from datetime import datetime
from subprocess import PIPE, STDOUT, Popen

from .fileio import get_tmpdir


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def clone(name):
    """
    Clone a repository by name
    """
    tmpdir = get_tmpdir()
    res = run_command(
        ["git", "clone", "--depth", "1", f"https://github.com/{name}", tmpdir]
    )

    # Don't fail entire process, just can't get dependencies
    if res["return_code"] != 0:
        return
    return tmpdir


def run_command(cmd, sudo=False, stream=False):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    stdout = PIPE if not stream else None
    if sudo is True:
        cmd = ["sudo"] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=stdout)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")

    return output


def get_time_now():
    """
    Get a timestring for the datetime now.
    """
    return datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")


def confirm_action(question, force=False):
    """confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question)
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True


def confirm_uninstall(filename, force=False):
    """confirm if the user wants to uninstall a module

    Parameters
    ==========
    filename: the file that will be removed
    force: if the user wants to skip the prompt
    """
    return confirm_action(
        "Are you sure you want to uninstall {}".format(filename), force
    )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

schema_url = "http://json-schema.org/draft-07/schema"


# The simplest form of aliases is key/value pairs
keyvals = {
    "type": "object",
    "patternProperties": {
        "\\w[\\w-]*": {"type": "string"},
    },
}


versionProperties = {
    "number": {"type": ["string", "number"]},
    "published_at": {"type": ["null", "string"]},
    "spdx_expression": {"type": ["null", "string"]},
    "original_license": {"type": ["null", "string"]},
    "researched_at": {"type": ["null", "string"]},
    "repository_sources": {"type": "array", "items": {"type": "string"}},
}

versions = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["number"],
        "properties": versionProperties,
    },
}


# This matches a libraries.io package
packageProperties = {
    "dependent_repos_count": {"type": "number"},
    "dependents_count": {"type": "number"},
    "deprecation_reason": {"type": ["null", "string"]},
    "description": {"type": ["null", "string"]},
    "forks": {"type": "number"},
    "homepage": {"type": ["null", "string"]},
    "keywords": {"type": "array", "items": {"type": "string"}},
    "language": {"type": ["null", "string"]},
    "latest_download_url": {"type": ["null", "string"]},
    "latest_release_number": {"type": ["null", "string"]},
    "latest_release_published_at": {"type": ["null", "string"]},
    "latest_stable_release_number": {"type": ["null", "string"]},
    "latest_stable_release_published_at": {"type": ["null", "string"]},
    "license_normalized": {"type": "boolean"},
    "licenses": {"type": ["null", "string"]},
    "name": {"type": "string"},
    "normalized_licenses": {"type": "array", "items": {"type": "string"}},
    "package_manager_url": {"type": ["null", "string"]},
    "platform": {"type": ["null", "string"]},
    "rank": {"type": ["null", "number"]},
    "repository_license": {"type": ["null", "string"]},
    "repository_url": {"type": ["null", "string"]},
    "stars": {"type": ["null", "number"]},
    "status": {"type": ["null", "string"]},
    "versions": versions,
}

# We use libraries.io schemas to ensure consistency,
# and we only require what we absolutely need!
package = {
    "$schema": schema_url,
    "title": "Package Schema",
    "type": "object",
    "required": ["description", "homepage", "language", "name", "versions"],
    "properties": packageProperties,
}

# Currently all of these are required
settingsProperties = {
    "cache_dir": {"type": "string"},
    "disable_cache": {"type": "boolean"},
    "disable_memory_cache": {"type": "boolean"},
}

settings = {
    "$schema": schema_url,
    "title": "Settings Schema",
    "type": "object",
    "required": [
        "disable_memory_cache",
        "disable_cache",
        "cache_dir",
    ],
    "properties": settingsProperties,
    "additionalProperties": False,
}



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.cache as cache
import citelang.main.endpoints as endpoints
import citelang.main.packages as packages
from citelang.logger import logger


def get_package(
    manager,
    name,
    version=None,
    data=None,
    use_cache=True,
    manager_kwargs=None,
):
    """
    Return a package handle for a custom package (defined in citelang) or libraries.io
    """
    # Custom kwargs for the manager
    manager_kwargs = manager_kwargs or {}

    # Case 1: a package manager provided by citelang
    if manager in packages.manager_names:
        return CustomPackage(
            name=name,
            manager=manager,
            version=version,
            data=data,
            use_cache=use_cache,
            manager_kwargs=manager_kwargs,
        )

    # Case 2: libraries.io
    return Package(
        name=name,
        manager=manager,
        version=version,
        data=data,
        use_cache=use_cache,
    )


class PackageBase:
    def __init__(
        self,
        manager,
        name,
        version=None,
        data=None,
        use_cache=True,
        manager_kwargs=None,
    ):
        """
        The version can be set explicitly or come from the name.
        """
        self.manager = manager
        self.version = version
        self.parse(name)
        self.data = data or {}
        self.latest = None
        self.use_cache = use_cache
        self.cache = cache.cache
        self.manager_kwargs = manager_kwargs or {}
        self.underlying_manager = None

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[citelang-package]"

    @property
    def cache_name(self):
        return f"package/{self.manager}/{self.name}"

    @property
    def homepage(self):
        url = self.data.get("package", {}).get("homepage")
        if not url and self.manager == "pypi":
            return f"https://pypi.org/project/{self.name}"
        if not url and self.manager == "go":
            return f"https://{self.name}"
        # Some R urls have TWO and they are wonky
        if url and " " in url:
            url = url.strip().split(" ")[0]
        return url

    def parse(self, name):
        """
        Parse a manager and name into the package
        """
        if "@" in name:
            name, version = name.split("@", 1)
            self.version = version
        if "[" in name and "]" in name:
            name = name.split("[", 1)[0]

        # invalid characters found!
        name = name.replace(";", "")
        self.name = name

    def dependencies(self, return_data=False):
        raise NotImplementedError

    def info(self):
        raise NotImplementedError


class CustomPackage(PackageBase):
    """
    A wrapper for a custom package (provided by citelang)
    """

    def __init__(
        self,
        manager,
        name,
        version=None,
        data=None,
        use_cache=True,
        manager_kwargs=None,
    ):
        super().__init__(manager, name, version, data, use_cache, manager_kwargs)
        if not self.underlying_manager:
            self.underlying_manager = packages.managers[self.manager](
                **self.manager_kwargs
            )

    @property
    def homepage(self):
        url = self.data.get("homepage") or self.data.get("package", {}).get("homepage")
        if not url and self.manager == "github":
            return f"https://github.com/{self.name}"
        if not url and self.manager in ["go.mod", "go"]:
            return f"https://{self.name}"
        if url and " " in url:
            url = url.strip().split(" ")[0]
        return url

    def dependencies(self, return_data=False):
        """
        Retrieve custom package dependencies
        """
        manager = self.underlying_manager

        # Use manager cache first
        if "dependencies" in manager.data:
            return manager.data["dependencies"]

        cache_name = None
        result = None

        # If we aren't given a version, look for default branches
        if not self.version and manager.default_versions and self.use_cache:
            for version in manager.default_versions:
                cache_name = f"package/{self.manager}/{self.name}/{version}"
                result = self.cache.get(cache_name)
                if result:
                    result = endpoints.get_endpoint("dependencies", data=result)
                    self.version = version
                    break

        # If we don't have a version, we have to retrieve an update
        if not result and not self.version:
            deps = manager.package(self.name)
            result = endpoints.get_endpoint("dependencies", data=deps)
            self.version = result.data.get("default_version")
            cache_name = f"package/{self.manager}/{self.name}/{self.version}"

        # We have a version, either retrieve from cache or anew
        elif not result and self.version:
            cache_name = f"package/{self.manager}/{self.name}/{self.version}"
            result = self.cache.get(cache_name)
            if not result or not self.use_cache:
                version = "@%s" % self.version if self.version else ""
                logger.info("Retrieving new result for %s%s..." % (self.name, version))
                deps = manager.package(self.name)
                result = endpoints.get_endpoint("dependencies", data=deps)
            elif result:
                result = endpoints.get_endpoint("dependencies", data=result)

        # If cache is enabled, we save the result
        if result and cache_name:
            self.cache.set(cache_name, result)

        # Return the wrapped result or raw data
        deps = result.data.get("dependencies", [])
        self.data["package"] = result.data
        self.data["dependencies"] = deps
        if return_data:
            return deps
        return result

    def info(self):
        """
        Get info for a custom package
        """
        manager = self.underlying_manager

        # Use manager cache first
        if "package" in manager.data:
            return manager.data["package"]

        # First try retrieving from the cache
        result = self.cache.get(self.cache_name)
        if not result or not self.use_cache:
            version = "@%s" % self.version if self.version else ""
            logger.info(
                "Retrieving new result for package %s%s..." % (self.name, version)
            )
            pkg = manager.package(self.name)
            result = endpoints.get_endpoint(
                "package", data=pkg, package_name=self.name, manager=self.manager
            )
        else:
            result = endpoints.get_endpoint(
                "package", data=result, package_name=self.name, manager=self.manager
            )

        # If cache is enabled, we save the result
        self.cache.set(self.cache_name, result)
        self.data["package"] = result.data
        return result


class Package(PackageBase):
    """
    A basic wrapper for package parsing functions (libraries.io)
    """

    def dependencies(self, return_data=False):
        """
        Retrieve libraries.io dependencies
        """
        if not self.version:
            self.info()
            self.version = self.latest

        # Check for result in cache
        cache_name = f"dependencies/{self.manager}/{self.name}/{self.version}"
        result = self.cache.get(cache_name)
        if not result or not self.use_cache:
            logger.info(
                "Retrieving new result for %s@%s dependencies..."
                % (self.name, self.version)
            )
            result = endpoints.get_endpoint(
                "dependencies",
                manager=self.manager,
                package_name=self.name,
                version=self.version,
            )
        else:
            result = endpoints.get_endpoint("dependencies", data=result)

        # If cache is enabled, we save the result
        self.cache.set(cache_name, result)
        deps = result.data.get("dependencies", [])
        self.data["dependencies"] = deps
        if return_data:
            return deps
        return result

    def info(self):
        """
        Get info for a libraries.io package
        """
        result = self.cache.get(self.cache_name)
        if not result and not self.manager:
            return result

        if not result or not self.use_cache:
            version = "@%s" % self.version if self.version else ""
            logger.info(
                "Retrieving new result for package %s%s..." % (self.name, version)
            )
            result = endpoints.get_endpoint(
                "package",
                manager=self.manager,
                package_name=self.name,
            )
        else:
            result = endpoints.get_endpoint(
                "package", data=result, manager=self.manager, package_name=self.name
            )

        # Set a latest version if we find one
        if "versions" in result.data and result.data["versions"]:
            self.latest = result.data["versions"][-1]["number"]
        self.cache.set(self.cache_name, result)
        self.data["package"] = result.data
        return result



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

import os

import requests

import citelang.utils as utils
from citelang.logger import logger

from .base import PackageManager


def find_dependencies(repo):
    """
    Given a repository root, look for dependency files. We could use the GraphQL
    library, but this REQUIRES a GitHub token, which is kind of annoying. For now
    we are trying to scrape them :)
    """
    try:
        from bs4 import BeautifulSoup
    except ImportError:
        logger.exit(
            "You need beautiful soup in order to get GitHub dependencies. pip install bs4"
        )

    url = "https://github.com/{}/network/dependencies".format(repo)

    repos = set()
    while url is not None:
        response = requests.get(url)
        if response.status_code != 200:
            logger.warning("Issue getting dependencies for %s" % repo)
            return [{"name": x} for x in sorted(list(repos))]
        soup = BeautifulSoup(response.content, "html.parser")

        for t in soup.findAll("div", {"class": "Box-row"}):
            repo = t.find("a", {"data-hovercard-type": "repository"})
            if repo:
                repo = repo.text.replace(" ", "").strip()
                repos.add(repo)

        # Do we have pagination?
        pagination = soup.find("div", {"class": "paginate-container"})
        url = None
        if pagination:
            pagination_a = pagination.find("a")
            if pagination_a:
                url = pagination_a["href"]

    # Ensure dependencies come with homepage
    return [
        {"name": x, "homepage": "https://github.com/%s" % x}
        for x in sorted(list(repos))
    ]


class GitHubManager(PackageManager):
    """
    Packages from GitHub, either release, or branch.
    """

    name = "github"
    apiroot = "https://api.github.com"
    homepage = "https://github.com"
    color = "#000000"
    default_language = None
    project_count = None
    default_versions = ["main", "master", "develop"]

    def clone(self, name):
        """
        Clone a repository and sniff for dependency files.
        """
        return utils.clone(name)

    def package(self, name, **kwargs):

        # Ensure name in right format
        if not name.count("/") == 1:
            logger.exit("Please provide a repository in the format <org>/<name")

        headers = {
            "Accept": "application/vnd.github.v3+json;application/vnd.github.antiope-preview+json;application/vnd.github.shadow-cat-preview+json"
        }
        token = os.environ.get("GITHUB_TOKEN")
        if token:
            headers["Authorization"] = "token %s" % token
        else:
            logger.warning("Your API requests will be limited without GITHUB_TOKEN")

        repo = self.get_or_fail(f"{self.apiroot}/repos/{name}", headers)

        # Try to provide a default version
        repo["default_version"] = repo["default_branch"]

        # Parse repos dependency page. This includes deps for CI too.
        repo["dependencies"] = find_dependencies(name)
        repo["name"] = repo["full_name"]

        # Get versions from releases API
        releases = self.get_or_fail(
            f"{self.apiroot}/repos/{name}/releases?per_page=100", headers
        )
        repo["versions"] = [
            {"published_at": x["published_at"], "number": x["tag_name"]}
            for x in releases
        ]
        if not repo["homepage"]:
            repo["homepage"] = repo["html_url"]
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import json
import re

from .base import PackagesFromFile


class NPMPackageManager(PackagesFromFile):
    """
    Packages from a package.json
    """

    name = "package.json"
    underlying_manager = "npm"
    default_language = "JavaScript"
    project_count = None
    homepage = "https://www.npmjs.com/"
    color = "#d54d25"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Parse the self.content (the packages.json)
        """
        repo = self.get_repo()

        try:
            meta = json.loads(content)
        except Exception:
            meta = {}

        deps = []
        for package_name, version in meta.get("dependencies", {}).items():

            version = re.sub("(\^|<|>|=)", "", version)  # noqa

            # Always get rid of @ - doesn't seem to get hits
            package_name = package_name.replace("@", "", 1)
            pkg = self.get_package(package_name, version)
            if not package_name:
                continue

            # If we don't have a package, try using start of name
            if not pkg:
                package_name = package_name.split("/")[0]
                pkg = self.get_package(package_name, version)
                if not pkg:
                    continue

            if not pkg or not package_name:
                continue

            # Ensure we have version, fallback to latest
            if not version:
                version = pkg.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/npm/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": pkg.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["NPM"],
            }
            deps.append(dep)

        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

from .cpp_cmakelist import CMakeListManager
from .cran_description import RPackageManager
from .github import GitHubManager
from .go_mod import GoModuleManager
from .npm_packages import NPMPackageManager
from .pypi_requirements import RequirementsManager, SetupManager
from .ruby_gem import GemfileManager
from .spack import SpackManager

# Registered endpoints (populated on init)
managers = {}
manager_names = []
filesystem_manager_names = []

for manager in [
    GitHubManager,
    SpackManager,
    RequirementsManager,
    RPackageManager,
    SetupManager,
    GemfileManager,
    GoModuleManager,
    NPMPackageManager,
    CMakeListManager,
]:
    manager_names.append(manager.name)
    managers[manager.name] = manager

    if manager.filesystem_manager:
        filesystem_manager_names.append(manager.name)

manager_names.sort()
filesystem_manager_names.sort()



#!/usr/bin/env python3

import glob
import os

from pybind11.setup_helpers import ParallelCompile, Pybind11Extension
from setuptools import setup

# see https://github.com/pybind/python_example


def readme():
    with open("README.md") as f:
        return f.read()


def list_files(dirs, exts, exclude=[]):
    files = []
    if isinstance(exclude, str):
        exclude = [exclude]
    for directory in dirs:
        for ext in exts:
            files.extend(glob.glob(os.path.join(directory, "*." + ext)))
    [f in files and files.remove(f) for f in exclude]
    return files


# def cxxflags(platform):
#   if platform == "unix":
#     return [
#       "-Wall",
#       "-pedantic",
#       "-pthread",
#       "-Wsign-compare",
#       "-fstack-protector-strong",
#       "-Wformat",
#       "-Werror=format-security",
#       "-Wdate-time",
#       "-fPIC",
#       "-std=c++17",
#       "-fvisibility=hidden"
#     ]
#   elif platform == "msvc":
#     return ['/EHsc']
#   else:
#     return []

# def ldflags(_platform):
#   return []


ext_modules = [
    Pybind11Extension(
        "_neworder_core",
        sources=list_files(["src"], ["cpp"]),
        depends=["setup.py", "neworder/__init__.py"] + list_files(["src"], ["h"]),
        cxx_std=17,
    ),
]

ParallelCompile().install()

setup(
    name="neworder",
    # version set from __init__.py via setup.cfg,
    author="Andrew P Smith",
    author_email="a.p.smith@leeds.ac.uk",
    url="https://neworder.readthedocs.io",
    description="A dynamic microsimulation framework",
    long_description=readme(),
    long_description_content_type="text/markdown",
    packages=["neworder"],
    package_data={"neworder": ["py.typed", "*.pyi"]},
    ext_modules=ext_modules,
    install_requires=["numpy>=1.19.1", "pandas>=1.0.5", "scipy"],
    setup_requires=["pybind11>=2.5.0", "pytest-runner"],
    tests_require=["pytest", "mpi4py>=3.0.3"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.base as base
import citelang.main.parser as parser
from citelang.main.contrib import ContributionParser

Client = base.BaseClient
assert ContributionParser


def get_parser(filename=None, *args, **kwargs):
    """
    Get a parser based on a filename. A GraphClient is required.
    """
    if not filename:
        return parser.FileNameParser(*args, **kwargs)
    if filename.endswith(".md"):
        return parser.FileNameParser(filename=filename, *args, **kwargs)
    return parser.RequirementsParser(filename=filename, *args, **kwargs)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import multiprocessing
import os
import re

import citelang.utils as utils
from citelang.logger import logger, spinner

# global pool used by parser functions
pool = None


# Patterns to parse git blame
patterns = {
    "commit": "^([0-9a-f]{40})",
    "author": "^author (.*)$",
    "text": "^\t(.*)$",
    "time": "^committer-time (.*)$",
}


def blame_task(args):
    """
    Run the blame task. This is based on logic from pypi contrib.
    If there is an empty result, meaning no commits match or there is an issue
    with blame, record the file as empty in the cache so we don't try again.
    """
    # Sploot out args into variables
    root, commit, path, save_to, shallow = args

    # If we have the history already, return
    if os.path.exists(save_to):
        return {
            "output": utils.read_json(save_to),
            "path": path,
            "exist": True,
            "empty": "",
        }

    # Atomic save
    tmp_file = save_to + ".tmp.%d" % os.getpid()

    # Asssemble command
    command = ["git", "blame", "-w", "--line-porcelain"]

    # find line copies and movements across files (more shallow result)
    if shallow:
        command += ["-M", "-C"]
    command += [commit, "--", path]
    with utils.workdir(root):
        try:
            res = utils.run_command(command)
        # This means we hit some binary, etc.
        except Exception:
            return {"empty": "%s-%s" % (commit, path)}

    # Do not continue if there is any error
    if res["return_code"] != 0:
        return {"empty": "%s-%s" % (commit, path)}

    # This is the raw git blame
    items = parse_blame_output(res["message"].strip(), path)
    if not items:
        return {"empty": "%s-%s" % (commit, path)}

    # Structure in cache mirrors structure of repo
    utils.mkdir_p(os.path.dirname(tmp_file))
    utils.write_json(items, tmp_file)
    os.rename(tmp_file, save_to)
    return {"output": items, "path": path, "exist": False, "empty": ""}


def parse_blame_output(output, path):
    """
    Helper function to parse blame output and return list.
    """
    # Search each line in output, derive commits, authors, and text
    items = {}
    item = {}
    for line in output.split("\n"):

        for pattern_type, pattern in patterns.items():
            match = re.search(pattern, line)
            if match:
                item[pattern_type] = match.group(1)

        # We've parsed a single item - save complete time information etc
        if all(key in item for key in ("commit", "author", "text", "time")):

            # Don't include empty lines
            if item["text"].strip() == "":
                item = {}
                continue
            item.update({"path": path})

            # We don't convert time to int here because it gets loaded as str
            current = items
            for key in ["author", "path", "commit"]:
                if item[key] not in current:
                    current[item[key]] = {}
                current = current[item[key]]

            if item["time"] not in current:
                current[item["time"]] = 0
            current[item["time"]] += 1
            item = {}

    return items


class GitParser:
    """
    Shared functions to parse git repos, etc.
    """

    def git(self, *cmd):
        """
        Run a git command (in context of the root)
        """
        with utils.workdir(self.root):
            res = utils.run_command(cmd)
            if res["return_code"] != 0:
                logger.exit(
                    "Issue with command %s: %s" % (" ".join(cmd), res["message"])
                )
            return res["message"].strip()


class CommitStats(GitParser):
    """
    Stats for a particular commit, saved to a cache.
    """

    def __init__(self, root, commit, paths, outdir, shallow=False):
        self.paths = paths
        self.commit = commit
        self.items = {}
        self.root = root
        self.shallow = shallow

        # This is a shared cache of blame output
        self.outdir = outdir
        self.tasks = ()
        self.prepare()

    @property
    def ntasks(self):
        return len(self.tasks)

    @property
    def empty_file(self):
        return os.path.join(self.outdir, ".citelang-empty-results.txt")

    def prepare(self):
        """
        Prepare the list of things to parse. We do this to get a count in advance.
        """
        # We need to take intersection of files for commit and those in paths
        files = self.git(
            "git", "ls-tree", "-r", "--name-only", "--full-tree", self.commit
        ).split("\n")

        if self.paths:
            files = [x for x in files if x in self.paths]

        # A listing of known empty commit-path combos
        self.empties = set()
        if os.path.exists(self.empty_file):
            [
                self.empties.add(x.strip())
                for x in utils.read_file(self.empty_file).split("\n")
            ]

        # Prepare args for imap unordered pool
        self.tasks = [
            (
                self.root,
                self.commit,
                f,
                os.path.join(self.outdir, f, f"{self.commit}.json"),
                self.shallow,
            )
            for f in set(files)
            if "%s-%s" % (self.commit, f) not in self.empties
        ]

    def run(self, pool, complete, total):
        """
        Run git blame to derive stats for the commit and paths, use pool.
        """
        seen = set()
        results = pool.imap_unordered(blame_task, self.tasks)
        for result in results:
            pad = utils.get_terminal_pad(25)
            print("Done {:.2f} %".format(complete / total * 100).ljust(pad), end="\r")
            complete += 1

            # No result - record to empty so we don't try again
            empty = result["empty"]
            if empty:
                self.empties.add(empty)
                continue

            output = result["output"]
            path = result["path"]

            if path not in seen and not result["exist"]:
                seen.add(path)

            # This includes all commits found in the blame over time
            self.items = utils.update_nested(self.items, output)

        # Save empties to file
        utils.write_file("\n".join(list(self.empties)), self.empty_file)


class ContributionSummary:
    """
    A summary of contributions by author, line, file, etc.
    """

    def __init__(self, history, start_timestamp=None, end_timestamp=None):
        self.history = history

        # If we are filtering git blame to within a range
        self.start_timestamp = start_timestamp
        self.end_timestamp = end_timestamp

    def iter_items(self, filters=None):
        """
        Yield authors and paths from the history, ensuring that a duplicate
        result from a git blame (a commit persisting across time) does not get
        counted twice.
        """
        filters = utils.read_yaml(filters) if filters else {}
        authors = filters.get("authors", {})
        ignore_files = filters.get("ignore_files", [])
        ignore_basename = filters.get("ignore_basename", [])
        ignore_users = filters.get("ignore_users", [])
        ignore_bots = filters.get("ignore_bots", False)

        seen = set()
        for commit, items in self.history.items():
            for author, paths in items.items():
                for path, commits in paths.items():

                    # ignore specific paths
                    if (
                        os.path.basename(path) in ignore_basename
                        or path in ignore_files
                    ):
                        continue

                    for commit, times in commits.items():
                        for timestamp, count in times.items():

                            # **important** do not count a contribution more than once
                            timestamp = int(timestamp)
                            uid = (commit, author, path, timestamp)
                            if uid in seen:
                                continue
                            seen.add(uid)

                            # Don't include this commit, too early
                            if (
                                self.start_timestamp
                                and int(timestamp) < self.start_timestamp
                            ):
                                continue

                            # Or too late!
                            if (
                                self.end_timestamp
                                and int(timestamp) > self.end_timestamp
                            ):
                                continue

                            # Apply filters here
                            original_author = author
                            if author in authors:
                                author = authors[author]

                            # Is it a bot or do we choose to ignore?
                            if (
                                ignore_bots
                                and "[bot]" in author
                                or author in ignore_users
                                or original_author in ignore_users
                            ):
                                continue

                            yield commit, author, path, int(timestamp), count

    def by_file(self, detail=True, filters=None):
        """
        Return summary of contributions by path
        """
        paths = {}
        for _, author, path, _, count in self.iter_items(filters):

            # Detailed returns results by author
            if detail:
                if path not in paths:
                    paths[path] = {}
                paths[path][author] = count

            # Or just total lines per path
            else:
                if path not in paths:
                    paths[path] = 0
                paths[path] += count

        return self._to_list(paths, detail)

    def by_author(self, detail=True, filters=None):
        """
        Return summary of contributions by author
        """
        results = {}
        for _, author, path, _, count in self.iter_items(filters):

            # Detailed returns results by filename
            if detail:
                if author not in results:
                    results[author] = {}
                results[author][path] = count

                # Or just total lines per author
            else:
                if author not in results:
                    results[author] = 0
                results[author] += count

        # Ensure we are sorted, and parse into list
        return self._to_list(results, detail)

    def _to_list(self, original, detail):
        """
        Convert an original dict (or nested dict) into a list.
        """
        items = []
        if detail:
            for author, paths in original.items():
                sorted_paths = {
                    k: v
                    for k, v in sorted(
                        paths.items(), key=lambda item: item[1], reverse=True
                    )
                }
                items.append({"name": author, "paths": sorted_paths})
        else:
            original = {
                k: v
                for k, v in sorted(
                    original.items(), key=lambda item: item[1], reverse=True
                )
            }
            for author, count in original.items():
                items.append({"name": author, "count": count})
        return items


class ContributionParser(GitParser):
    def __init__(self, root=None, start=None, end=None, outdir=None, paths=None):
        self.start = start or ""
        self.end = end or ""
        self.set_root(root)
        self.set_paths(paths)
        self.outdir = outdir or os.path.join(os.getcwd(), ".contrib")
        utils.mkdir_p(self.outdir)

    @property
    def git_root(self):
        return os.path.join(self.root, ".git")

    @property
    def name(self):
        """
        The repository name
        """
        return os.path.basename(self.root)

    def set_root(self, root):
        """
        Ensure the root exists and set on the class
        """
        root = os.path.abspath(root or os.getcwd())
        if not os.path.exists(root):
            logger.exit(f"{root} does not exist.")
        self.root = root

    def set_paths(self, prefixes=None):
        """
        Parse the repository and set list of paths to include
        """
        paths = set()
        prefixes = prefixes or []
        regex = "(%s)" % "|".join(prefixes) if prefixes else ""

        if not prefixes:
            self.paths = []
            return

        # Get all files in the commit history
        for path in self.git(
            "git", "log", "--pretty=format:", "--name-only", "--diff-filter=A"
        ).split("\n"):
            path = path.strip()
            if not path:
                continue
            if prefixes and re.search(regex, path):
                paths.add(path)
            elif not prefixes:
                paths.add(path)

        self.paths = list(paths)

    def parse(self, return_summary=True, within_range=True, shallow=False):
        """
        Parse the contributions. If within_range is True, don't include git
        blame that goes outside of the range provided.
        """
        # We must be in a git repository
        if not os.path.exists(self.git_root):
            logger.exit(f"Cannot find .git repo in {self.root}")

        # get commits associated with start to end
        commits = self.get_commit_range(shallow)

        # Retrieve items of history
        history = self.index_history(commits, shallow=shallow)

        # Do we want to only include within a range?
        start_timestamp = (
            None if not within_range else self.get_commit_timestamp(commits[0])
        )
        end_timestamp = (
            None if not within_range else self.get_commit_timestamp(commits[-1])
        )

        # Summarize lines by author
        if return_summary:
            return ContributionSummary(history, start_timestamp, end_timestamp)
        return history

    def get_commit_timestamp(self, commit):
        """
        Return the timestamp of a commit.
        """
        return int(self.git("git", "show", "-s", "--format=%ct", commit))

    def index_history(self, commits, shallow=False):
        """
        index git history.
        """
        # Keep a lookup of commit stats by directory
        parsers = {}

        jobs = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(jobs)

        # Create output cache directory just for blames
        cache = os.path.join(self.outdir, "cache")
        utils.mkdir_p(cache)

        # Cut out early if we have a result cached (based on start and end)
        if self.start and self.end:
            uid = "%s-%s" % (self.start, self.end)
            cached_result = os.path.join(self.outdir, "%s.json" % uid)
            if os.path.exists(cached_result):
                logger.info(f"Loading cached result {cached_result}")
                return utils.read_json(cached_result)

        spins = spinner()
        try:

            # First generatae parsers, count tasks in advance
            total = 0
            print("Preparing all tasks...", end="\r")
            for commit in commits:
                parser = CommitStats(
                    self.root, commit, self.paths, cache, shallow=shallow
                )
                print("Preparing all tasks...%s" % next(spins), end="\r")
                total += parser.ntasks
                parsers[commit] = parser

            # Now run parser which knowledge of total
            complete = 0
            print(f"Found {total} tasks!", end="\r")
            for commit, parser in parsers.items():
                parser.run(pool, complete, total)
                complete += parser.ntasks
        finally:
            pool.terminate()

        stats = {}
        for commit, parser in parsers.items():
            stats[commit] = parser.items

        # Cache the result
        if self.start and self.end:
            uid = "%s-%s" % (self.start, self.end)
            cached_result = os.path.join(self.outdir, "%s.json" % uid)
            utils.write_json(stats, cached_result)
        return stats

    def get_tag_commit(self, tag):
        """
        Given a tag, retrieve the commit for it.
        """
        return self.git("git", "rev-list", "-n", "1", tag)

    def get_commit_range(self, shallow=False):
        """
        Given a start and end, parse and return the commits from git
        """
        # Single commit, honor previous behavior
        if self.start and self.start == self.end:
            log_range = self.start + "~1.." + self.start
        # Open or closed range with start
        elif self.start:
            log_range = self.start + ".." + self.end
        # Open range with end (initial commit to end)
        elif self.end:
            log_range = self.end
        # No start or end... all history for HEAD
        else:
            log_range = "HEAD"

        if shallow:
            res = self.git("git", "log", "--first-parent", "--format=%H", log_range)

        # Possibly duplications but won't miss any commits
        else:
            res = self.git("git", "log", "--format=%H", log_range)

        # Get commits and timestamps
        commits = [x for x in res.split("\n") if x]

        # No commits?
        if not commits:
            logger.exit("This repository does not have any history.")

        # Need to reverse - end (most recent) is at top!
        commits.reverse()

        logger.info("Found %s commits." % len(commits))
        return commits



#!/usr/bin/python

import json
import os
import shutil

import pytest

import citelang.utils as utils


def test_write_read_files(tmp_path):
    """
    test_write_read_files will test the functions write_file and read_file
    """
    print("Testing utils.write_file...")

    tmpfile = str(tmp_path / "written_file.txt")
    assert not os.path.exists(tmpfile)
    utils.write_file("hello!", tmpfile)
    assert os.path.exists(tmpfile)

    print("Testing utils.read_file...")

    content = utils.read_file(tmpfile)
    assert content == "hello!"


def test_write_bad_json(tmp_path):

    bad_json = {"Wakkawakkawakka'}": [{True}, "2", 3]}
    tmpfile = str(tmp_path / "json_file.txt")
    assert not os.path.exists(tmpfile)
    with pytest.raises(TypeError):
        utils.write_json(bad_json, tmpfile)


def test_write_json(tmp_path):

    good_json = {"Wakkawakkawakka": [True, "2", 3]}
    tmpfile = str(tmp_path / "good_json_file.txt")

    assert not os.path.exists(tmpfile)
    utils.write_json(good_json, tmpfile)
    with open(tmpfile, "r") as f:
        content = json.loads(f.read())
    assert isinstance(content, dict)
    assert "Wakkawakkawakka" in content
    content = utils.read_json(tmpfile)
    assert "Wakkawakkawakka" in content


def test_get_installdir():
    print("Testing utils.get_installdir")

    whereami = utils.get_installdir()
    print(whereami)
    assert whereami.endswith("citelang")


def test_run_command():
    print("Testing utils.run_command")

    result = utils.run_command(["echo", "hello"])
    assert result["message"] == "hello\n"
    assert result["return_code"] == 0


def test_get_tmpdir_tmpfile():
    print("Testing utils.get_tmpdir, get_tmpfile")

    tmpdir = utils.get_tmpdir()
    assert os.path.exists(tmpdir)
    assert os.path.basename(tmpdir).startswith("citelang")
    shutil.rmtree(tmpdir)
    tmpdir = utils.get_tmpdir(prefix="name")
    assert os.path.basename(tmpdir).startswith("name")
    shutil.rmtree(tmpdir)
    tmpfile = utils.get_tmpfile()
    assert "citelang" in tmpfile
    os.remove(tmpfile)
    tmpfile = utils.get_tmpfile(prefix="pancakes")
    assert "pancakes" in tmpfile
    os.remove(tmpfile)


def test_mkdir_p(tmp_path):
    print("Testing utils.mkdir_p")

    dirname = str(tmp_path / "input")
    result = os.path.join(dirname, "level1", "level2", "level3")
    utils.mkdir_p(result)
    assert os.path.exists(result)


def test_print_json():
    print("Testing utils.print_json")

    result = utils.print_json({1: 1})
    assert result == '{\n    "1": 1\n}'



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import citelang.main.client as client
import citelang.utils as utils


def main(args, parser, extra, subparser):

    # Case 1: only one thing provided to package, and exists as a file
    if os.path.isfile(args.package[1]):
        cli = client.get_parser(filename=args.package[1], quiet=args.quiet)
        result = cli.gen(
            name=args.package[0],
            use_cache=not args.no_cache,
            max_depth=args.max_depth,
            max_deps=args.max_depth,
            min_credit=args.min_credit,
            credit_split=args.credit_split,
        )

    # Case 2: named package and manager
    else:
        cli = client.get_parser()
        result = cli.gen(
            name=args.package[1],
            manager=args.package[0],
            use_cache=not args.no_cache,
            max_depth=args.max_depth,
            max_deps=args.max_depth,
            min_credit=args.min_credit,
            credit_split=args.credit_split,
        )

    if args.outfile:
        utils.write_file(result.render(), args.outfile)
        print("Saved to %s" % args.outfile)
    else:
        print(result.render())



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import sys
from datetime import datetime

from citelang.logger import logger

from .base import GraphBase

template = """<?xml version="1.0" encoding="UTF-8"?>
<gexf xmlns="http://www.gexf.net/1.1draft" version="1.1" xmlns:viz="http://www.gexf.net/1.1draft/viz" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.gexf.net/1.1draft http://www.gexf.net/1.1draft/gexf.xsd">
    <meta lastmodifieddate="%s">
        <creator>citelang</creator>
        <description>Citation credit analysis of %s</description>
    </meta>
    <graph defaultedgetype="directed" idtype="string" type="static">
    <nodes>
"""


class Gexf(GraphBase):
    def generate(self):
        if self.outfile != sys.stdout:
            logger.info("Output will be written to %s" % self.outfile)
            fd = open(self.outfile, "w")
        else:
            fd = self.outfile

        today = datetime.now().strftime("%Y-%m-%d")

        # Add the binary name and date to the template
        fd.write(template % (today, self.data.name))

        # Create graph with nodes
        next_nodes = [self.data]
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            fd.write('        <node id="%s" label="%s"/>\n' % (label, name))
            for child in next_node.children:
                next_nodes.append(child)

        fd.write("    </nodes>\n    <edges>")

        # Now create links
        next_nodes = [self.data]
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            for child in next_node.children:
                label_to = self.uids[child.name]
                fd.write(
                    '        <edge id="%s" source="%s" target="%s" label="depends on"/>\n'
                    % (label_to, label, label_to)
                )
                next_nodes.append(child)

        fd.write("    </edges>\n")
        fd.write("</graph>\n")
        fd.write("</gexf>\n")

        if self.outfile != sys.stdout:
            fd.close()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import random

import colour

# Palettes

# https://coolors.co/a80874-ac4697-b083b9-b4c0dc-b7fdfe-8bf8c5-5ef38c-45c556-2b9720-343a1a
palettes = {
    "radish": [
        "a80874",
        "ac4697",
        "b083b9",
        "b4c0dc",
        "b7fdfe",
        "8bf8c5",
        "5ef38c",
        "45c556",
        "2b9720",
        "343a1a",
    ]
}


# Generally safe colors for light or black backgrounds in the 16 color palette.
# See also https://robotmoon.com/256-colors/
termColors = [
    1,
    2,
    3,
    4,
    5,
    6,
    9,
    10,
    11,
    12,
    13,
    14,
]


def get_random_color():
    return "#" + "".join([random.choice("ABCDEF0123456789") for i in range(6)])


def get_rich_color():
    """
    Return a random color
    """
    return "color(" + str(random.choice(termColors)) + ")"


def get_rich_colors(N):
    """
    Randomly generate N colors for rich (integers)
    """
    chosen_colors = []

    for i in range(N):
        color = None
        while not color or color in chosen_colors:
            color = get_rich_color()
        chosen_colors.append(color)
    return chosen_colors


def get_random_palette():
    """
    Choose a random color palette
    """
    name = random.choice(list(palettes.keys()))
    return palettes[name]


def get_color_range(N=100):
    """
    Get a color range, size N
    """
    start_color = colour.Color(get_random_color())
    end_color = colour.Color(get_random_color())
    return list(start_color.range_to(end_color, N))



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

import os
import sys

import requests

import citelang.main.cache as cache
import citelang.main.endpoints as endpoints
import citelang.utils as utils
from citelang.logger import logger


class PackageManager:
    """
    A package manager is a custom class (not supported by libraries io)
    """

    underlying_manager = None
    filesystem_manager = False

    def __init__(self, *args, **kwargs):
        self.cache = cache.cache
        self.data = {}
        for attr in [
            "name",
            "project_count",
            "homepage",
            "color",
            "default_language",
            "default_versions",
        ]:
            if not hasattr(self, attr):
                sys.exit(
                    "Misconfigured package manager %s missing %s attribute"
                    % (self.name, attr)
                )

    # If needed, make sure endpoint data is sorted
    def info(self):
        return {
            "name": self.underlying_manager or self.name,
            "project_count": self.project_count,
            "homepage": self.homepage,
            "color": self.color,
            "default_language": self.default_language,
        }

    def get_or_fail(self, url, headers=None, return_text=False):
        """
        Shared endpoint to get a url or fail.
        """
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            logger.exit("Cannot retrieve %s: %s" % (url, response.json()))
        if return_text:
            return response.text
        return response.json()

    def package(self, name, **kwargs):
        raise NotImplementedError


class PackagesFromFile(PackageManager):
    """
    Load packages from file. The class requires a parse function to parse
    the content provided.
    """

    filesystem_manager = True

    def __init__(self, package_name=None, content=None, filename=None):
        """
        An base filesystem package manager parses packages from a file
        """
        self.version = None
        self.package_name = package_name
        super().__init__()
        if package_name:
            self.set_name(package_name)
        self.data = {}
        self.filename = filename
        if content:
            self.parse(content)

    def set_name(self, name):
        if "@" in name:
            name, version = name.split("@", 1)
            self.version = version
        if "[" in name and "]" in name:
            name = name.split("[", 1)[0]

        # invalid characters found!
        name = name.replace(";", "")
        self.package_name = name

    def get_package(self, package_name, version=None):
        """
        Shared function to get a package based on name, version
        """
        # Try to get from cache - either versioned or not
        pkg = None
        if package_name and version:
            cache_name = f"package/{self.underlying_manager}/{package_name}/{version}"
            result = self.cache.get(cache_name)
            if result:
                pkg = endpoints.get_endpoint("package", data=result)

        elif package_name and not version:
            cache_name = f"package/{self.underlying_manager}/{package_name}"
            result = self.cache.get(cache_name)
            if result:
                pkg = endpoints.get_endpoint("package", data=result)

        if pkg is None:

            # Don't try again if this package is flagged as empty (not existing)
            if self.cache.is_empty(f"package/{self.underlying_manager}/{package_name}"):
                return pkg
            try:
                pkg = endpoints.get_endpoint(
                    "package",
                    package_name=package_name,
                    manager=self.underlying_manager,
                )
            except Exception:
                pass

        # If we get down here and no package, mark empty (unless it isn't a package we know)
        if not pkg:
            try:
                self.cache.mark_empty(
                    f"package/{self.underlying_manager}/{package_name}"
                )
            except Exception:
                pass
        return pkg

    def get_repo(self):
        """
        Helper function to return start of repo metadata. Intended to be used
        in the self.parse() function.
        """
        # The "repo" is the package name, we can't be sure about versions
        versions = [
            {
                "published_at": utils.get_time_now(),
                "number": self.version or "latest",
            }
        ]
        repo = {"name": self.package_name, "versions": versions}

        # Try to provide a default version
        repo["default_version"] = self.version or "latest"
        return repo

    def package(self, name, **kwargs):
        """
        The package endpoint ignores the name and just returns parsed data
        """
        return self.data.get("package")

    def from_file(self, filename, **kwargs):
        """
        From file reads in the filename and presents to parse.

        This is a means for a specific package manager to be used directly.
        We trust the user to provide the correct content file to parse.
        """
        self.filename = os.path.abspath(filename)
        content = utils.read_file(filename)
        return self.parse(content, **kwargs)

    def parse(self, content, **kwargs):
        """
        Parse the self.content
        """
        raise NotImplementedError



from .fileio import (
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    print_json,
    read_file,
    read_json,
    read_yaml,
    workdir,
    write_file,
    write_json,
)
from .string import get_terminal_pad, update_nested
from .terminal import (
    clone,
    confirm_action,
    confirm_uninstall,
    get_installdir,
    get_time_now,
    run_command,
)



import os

from setuptools import find_packages, setup


def get_lookup():
    lookup = dict()
    version_file = os.path.join("citelang", "version.py")
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup


def get_reqs(lookup=None, key="INSTALL_REQUIRES"):
    if lookup is None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup[key]:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" % (module_name, module_meta["exact_version"])
        elif "min_version" in module_meta:
            if module_meta["min_version"] is None:
                dependency = module_name
            else:
                dependency = "%s>=%s" % (module_name, module_meta["min_version"])
        install_requires.append(dependency)
    return install_requires


# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__))
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup["__version__"]
NAME = lookup["NAME"]
AUTHOR = lookup["AUTHOR"]
EMAIL = lookup["EMAIL"]
PACKAGE_URL = lookup["PACKAGE_URL"]
KEYWORDS = lookup["KEYWORDS"]
DESCRIPTION = lookup["DESCRIPTION"]
LICENSE = lookup["LICENSE"]

# Try to read description, otherwise fallback to short description
try:
    with open("README.md") as filey:
        LONG_DESCRIPTION = filey.read()
except Exception:
    LONG_DESCRIPTION = DESCRIPTION

################################################################################
# MAIN #########################################################################
################################################################################

if __name__ == "__main__":

    INSTALL_REQUIRES = get_reqs(lookup)
    TESTS_REQUIRES = get_reqs(lookup, "TESTS_REQUIRES")
    BADGE_REQUIRES = get_reqs(lookup, "BADGE_REQUIRES")
    INSTALL_REQUIRES_ALL = get_reqs(lookup, "INSTALL_REQUIRES_ALL")

    setup(
        name=NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=EMAIL,
        maintainer=AUTHOR,
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        url=PACKAGE_URL,
        license=LICENSE,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        keywords=KEYWORDS,
        setup_requires=["pytest-runner"],
        install_requires=INSTALL_REQUIRES,
        tests_require=TESTS_REQUIRES,
        extras_require={
            "all": [INSTALL_REQUIRES_ALL],
            "badge": [BADGE_REQUIRES],
        },
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Programming Language :: C",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3.8",
        ],
        entry_points={"console_scripts": ["citelang=citelang.client:run"]},
    )



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import argparse
import os
import sys

import citelang
from citelang.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="CiteLang",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--no-cache",
        dest="no_cache",
        help="disable using the cache for any particular command.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    description = "actions"
    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    # List package managers available
    ls = subparsers.add_parser(
        "list", description="list package managers available to derive citations from."
    )
    ls.add_argument("--json", help="output json", default=False, action="store_true")
    ls.add_argument("--all", help="output json", default=False, action="store_true")
    ls.add_argument("--outfile", "-o", help="write content to output file")

    # Cache control
    cache = subparsers.add_parser("cache", description="cache control")
    cache.add_argument(
        "--clear", help="clear the cache", default=False, action="store_true"
    )

    # Get a package or dependencies
    pkg = subparsers.add_parser(
        "package",
        description="list package managers available to derive citations from.",
    )
    deps = subparsers.add_parser("deps", description="list dependencies for a package.")

    # Extract contributions from a GitHub repo
    contrib = subparsers.add_parser(
        "contrib", description="Extract data for git contributions."
    )
    contrib.add_argument(
        "--outdir",
        "-d",
        help="output directory for data (defaults to .contrib)",
        default=os.path.join(os.getcwd(), ".contrib"),
    )
    contrib.add_argument("--outfile", "-o", help="Save to an output json file.")
    contrib.add_argument(
        "--start",
        "-s",
        help="start commit or tag to parse (defaults to start of history)",
    )
    contrib.add_argument(
        "--end",
        "-e",
        help="end commit or tag to parse (defaults to end of history)",
    )
    contrib.add_argument(
        "--root",
        "-r",
        help="Location where .git directory resides (defaults to PWD)",
        default=os.getcwd(),
    )
    contrib.add_argument(
        "--by",
        "-b",
        help="Organize results BY this attribute (author, path)",
        default="author",
        choices=["author", "path"],
    )
    contrib.add_argument(
        "--detail",
        help="include contribution details",
        default=False,
        action="store_true",
    )
    contrib.add_argument(
        "--shallow",
        help="Only look at the --first-parent and try to summarize copies and movements.",
        default=False,
        action="store_true",
    )
    contrib.add_argument(
        "--all-time",
        help="include blame from all time, not just within range.",
        default=False,
        action="store_true",
    )
    contrib.add_argument(
        "--filters",
        help="A filters yaml file, optionally with an authors key (key value pairs of an author name, and the alias to use) and an ignore key (list of author names to not include).",
    )

    badge = subparsers.add_parser(
        "badge", description="Generate an html (svg-based) badge."
    )
    badge.add_argument(
        "--template",
        "-t",
        help="Template to use for badge",
        choices=["treemap", "sunburst", "static"],
        default="static",
    )

    for command in [cache, badge]:
        command.add_argument(
            "--force",
            help="do not ask for confirmation",
            default=False,
            action="store_true",
        )

    graph = subparsers.add_parser(
        "graph", description="generate a graph for some package dependencies."
    )
    graph.add_argument(
        "--fmt", "-f", help="format of the graph (unset defaults to console print)"
    )
    credit = subparsers.add_parser(
        "credit", description="calculate dependency credit for a package."
    )
    gen = subparsers.add_parser(
        "gen", description="Generate a credit markdown file for a package of choice."
    )
    gen.add_argument("--outfile", "-o", help="Save to an output markdown file.")

    render = subparsers.add_parser(
        "render", description="render a credit tree into your markdown."
    )
    render.add_argument("filename", help="Markdown file to render software table into.")
    render.add_argument("--outfile", "-o", help="Save to an output json file.")

    for command in [pkg, deps, graph, credit, badge, gen]:
        command.add_argument(
            "package", help="package manager and name to parse", nargs=2
        )

    for command in [pkg, deps, graph, credit, badge]:
        command.add_argument(
            "--json",
            dest="json",
            help="print output to json.",
            default=False,
            action="store_true",
        )
        command.add_argument("--outfile", "-o", help="Save to an output json file.")

    for command in [graph, credit, badge, render, gen]:
        command.add_argument(
            "--max-depth", help="maximum depth to parse tree (default is unset)"
        )
        command.add_argument(
            "--max-deps",
            help="maximum number of dependencies to include (default is unset)",
        )
        command.add_argument(
            "--min-credit",
            help="minimum credit to allocate (and stop when lower)",
            type=float,
            default=0.01,
        )
        command.add_argument(
            "--credit-split",
            help="credit to allocate to main package and 1-split to dependents",
            type=float,
            default=0.5,
        )

    # Local shell with client loaded
    shell = subparsers.add_parser(
        "shell",
        description="shell into a Python session with a client.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    shell.add_argument(
        "--interpreter",
        "-i",
        dest="interpreter",
        help="python interpreter",
        choices=["ipython", "python", "bpython"],
        default="ipython",
    )

    # Custom config
    config = subparsers.add_parser(
        "config",
        description="update configuration settings. Use set or get to see or set information.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    config.add_argument(
        "--central",
        dest="central",
        help="make edits to the central config file.",
        default=False,
        action="store_true",
    )

    config.add_argument(
        "params",
        nargs="*",
        help="""Set or get a config value, edit the config, add or remove a list variable, or create a user-specific config.
citelang config set key:value
citelang config get key
citelang edit
citelang config inituser""",
        type=str,
    )
    return parser


def run():

    parser = get_parser()

    def help(return_code=0):
        version = citelang.__version__

        print("\nCiteLang Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(citelang.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "badge":
        from .badge import main
    elif args.command == "cache":
        from .cache import main
    elif args.command == "config":
        from .config import main
    elif args.command == "contrib":
        from .contrib import main
    elif args.command == "credit":
        from .credit import main
    elif args.command == "deps":
        from .deps import main
    elif args.command == "gen":
        from .gen import main
    elif args.command == "graph":
        from .graph import main
    elif args.command == "render":
        from .render import main
    elif args.command == "shell":
        from .shell import main
    elif args.command == "list":
        from .listing import main
    elif args.command == "package":
        from .package import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import collections.abc
import os


def get_terminal_pad(size):
    """
    Given an input of a particular size, get an estimated terminal left padding
    """
    try:
        # This will fail in a test environment
        return max(os.get_terminal_size().columns - size, 0)
    except Exception:
        return 30


def update_nested(data, update):
    """
    recursive function to update nested dict
    """
    for k, v in update.items():
        if isinstance(v, collections.abc.Mapping):
            data[k] = update_nested(data.get(k, {}), v)
        elif k not in data:
            data[k] = v
        elif (
            k in data
            and isinstance(data[k], (int, float))
            and isinstance(update[k], (int, float))
        ):
            data[k] = data[k] + data[v]
    return data



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import re

from citelang.logger import logger

from .base import PackagesFromFile


class PythonManager(PackagesFromFile):
    underlying_manager = "pypi"
    project_count = None
    homepage = "pypi.org/"
    color = "#006dad"
    default_language = "Python"
    default_versions = None

    def parse_python_deps(self, lines, **kwargs):
        """
        Shared function for deriving a list of python dependencies from lines
        """
        # Dependencies we parse as pypi packages
        # This should also update the cache and make it easier to retrieve later
        deps = []
        for line in lines:

            # Skip local installs with -e
            if "-e" in line:
                continue
            line = line.split("#", 1)[0]

            # skip comments
            if line.strip().startswith("#"):
                continue

            # We can't easily add github pypi references
            if not line or "git@" in line:
                continue

            version = None
            package_name = line.strip()
            if re.search("(==|~=|<=|>=|<|>|!=)", package_name):
                package_name, _, version = re.split(
                    "(==|~=|<=|>=|<|>|!=)", package_name
                )
                version = version.strip()
            package_name = package_name.strip()

            # We cannot parse a dep without a name
            if not package_name:
                continue

            # Remove variants
            package_name = re.sub("\[.+\]", "", package_name)  # noqa

            # First add requirements (names and pypi manager) to deps
            pkg = self.get_package(package_name, version)
            if not pkg:
                logger.warning("Issue getting package %s, skipping" % package_name)
                continue

            # Ensure we have version, fallback to latest
            if not version:
                version = pkg.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/pypi/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": pkg.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["Pypi"],
            }
            deps.append(dep)
        return deps
